/*      This is the LCDproc driver for Rackable Systems Roamer LCD.

        Copyright (C) 2008 Shad L. Lords

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2 of the License, or
        any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software
        Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301


        List of driver entry point:

        init            Implemented.
        close           Implemented.
        width           Implemented.
        height          Implemented.
        clear           Implemented by space filling no custom char info.
        flush           Implemented.
        string          Implemented.
        chr             Implemented.
        set_contrast    Implemented.
        backlight       Implemented.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "lcd.h"
#include "roamer.h"
#include "report.h"

// Variables
typedef struct {
    int fd;
    char lcd_init[3];
    char lcd_close[4];
    int width;
    int height;
    char *framebuf;
    char *backingstore;
    char beat;
} PrivateData;

// Vars for the server core
MODULE_EXPORT char *api_version = API_VERSION;
MODULE_EXPORT int stay_in_foreground = 0;
MODULE_EXPORT int supports_multiple = 1;
MODULE_EXPORT char *symbol_prefix = "roamer_";


MODULE_EXPORT int
roamer_init (Driver *drvthis)
{
    struct termios portset;

    char device[256] = DEFAULT_DEVICE;
    int speed = DEFAULT_SPEED;
    char size[256] = DEFAULT_SIZE;
    char buf[256] = "";
    int tmp, w, h;

    PrivateData *p;

    /* Alocate and store private data */
    p = (PrivateData *) malloc(sizeof(PrivateData));
    if (p == NULL)
        return -1;
    if (drvthis->store_private_ptr(drvthis, p))
        return -1;

    /* Initialise the PrivateData structure */
    p->fd = -1;

    p->width = LCD_DEFAULT_WIDTH;
    p->height = LCD_DEFAULT_HEIGHT;

    p->framebuf = NULL;
    p->backingstore = NULL;

    /* Control codes for display */
    snprintf(p->lcd_init, 3, "%c%c", 0x1e, 0x1d);
    snprintf(p->lcd_close, 4, "%c%c%c", 0x1e, 0x1b, 0x1b);

    /* Get serial device to use */
    strncpy(device, drvthis->config_get_string(drvthis->name, "Device", 0, DEFAULT_DEVICE), sizeof(device));
    device[sizeof(device)-1] = '\0';
    report(RPT_INFO, "%s: using Device %s", drvthis->name, device);

    /* Get display size */
    strncpy(size, drvthis->config_get_string(drvthis->name, "Size", 0, DEFAULT_SIZE), sizeof(size));
    size[sizeof(size)-1] = '\0';
    if ((sscanf(size, "%dx%d", &w, &h) != 2)
            || (w <= 0) || (w > LCD_MAX_WIDTH)
            || (h <= 0) || (h > LCD_MAX_HEIGHT)) {
        report(RPT_WARNING, "%s: cannot read Size: %s; using default %s",
                drvthis->name, size, DEFAULT_SIZE);
        sscanf(DEFAULT_SIZE , "%dx%d", &w, &h);
    }
    p->width = w;
    p->height = h;

    /* Get speed */
    tmp = drvthis->config_get_int(drvthis->name, "Speed", 0, DEFAULT_SPEED);
    switch (tmp) {
        case 1200:
            speed = B1200;
            break;
        case 2400:
            speed = B2400;
            break;
        case 4800:
            speed = B4800;
            break;
        case 9600:
            speed = B9600;
            break;
        default:
            speed = B9600;
            report(RPT_WARNING, "%s: Speed must be 1200, 2400, 4800 or 9600; using default %d",
                    drvthis->name, tmp);
    }

    /* Set up io port correctly, and open it... */
    p->fd = open(device, O_RDWR | O_NOCTTY);
    if (p->fd == -1) {
        report(RPT_ERR, "%s: open(%s) failed (%s)", drvthis->name, device, strerror(errno));
        if (errno == EACCES)
            report(RPT_ERR, "%s: %s device could not be opened...", drvthis->name, device);
        return -1;
    }
    report(RPT_INFO, "%s: opened display on %s", drvthis->name, device);

    tcgetattr(p->fd, &portset);

    // THIS ALL COMMENTED OUT BECAUSE WE NEED TO SET TIMEOUTS
    /* We use RAW mode */
#ifdef HAVE_CFMAKERAW_NOT
    /* The easy way */
    cfmakeraw(&portset);
#else
    /* The hard way */
    portset.c_iflag &= ~( IGNBRK | BRKINT | PARMRK | ISTRIP
            | INLCR | IGNCR | ICRNL | IXON );
    portset.c_oflag &= ~OPOST;
    portset.c_lflag &= ~( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
    portset.c_cflag &= ~( CSIZE | PARENB | CRTSCTS );
    portset.c_cflag |= CS8 | CREAD | CLOCAL;
    portset.c_cc[VMIN] = 1;
    portset.c_cc[VTIME] = 3;
#endif

    /* Set port speed */
    cfsetospeed(&portset, speed);
    cfsetispeed(&portset, B0);

    /* Do it... */
    tcsetattr(p->fd, TCSANOW, &portset);

    write(p->fd, p->lcd_init, sizeof(p->lcd_init)-1);

    /* Make sure the frame buffer is there... */
    p->framebuf = (unsigned char *) calloc(p->width * p->height, 1);
    if (p->framebuf == NULL) {
        report(RPT_ERR, "%s: unable to create framebuffer", drvthis->name);
        return -1;
    }
    memset(p->framebuf, ' ', p->width * p->height);

    /* make sure the framebuffer backing store is there... */
    p->backingstore = (unsigned char *) malloc(p->width * p->height);
    if (p->backingstore == NULL) {
        report(RPT_ERR, "%s: unable to create framebuffer backing store", drvthis->name);
        return -1;
    }
    memset(p->backingstore, ' ', p->width * p->height);

    /* set initial LCD configuration */
    roamer_clear(drvthis);

    report(RPT_DEBUG, "%s: init() done", drvthis->name);

    return 1;
}

/////////////////////////////////////////////////////////////////
// Closes the device
//
MODULE_EXPORT void
roamer_close (Driver *drvthis)
{
    PrivateData *p = drvthis->private_data;

    if (p != NULL) {
        write(p->fd, p->lcd_close, sizeof(p->lcd_close)-1);

        if (p->fd >= 0)
            close(p->fd);

        if (p->framebuf)
            free(p->framebuf);
        p->framebuf = NULL;

        if (p->backingstore)
            free(p->backingstore);
        p->backingstore = NULL;

        free(p);
    }
    drvthis->store_private_ptr(drvthis, NULL);
    report(RPT_DEBUG, "%s: closed", drvthis->name);
}

/////////////////////////////////////////////////////////////////
// Returns the display width
//
MODULE_EXPORT int
roamer_width (Driver *drvthis)
{
    PrivateData *p = drvthis->private_data;

    return p->width;
}

/////////////////////////////////////////////////////////////////
// Returns the display height
//
MODULE_EXPORT int
roamer_height (Driver *drvthis)
{
    PrivateData *p = drvthis->private_data;

    return p->height;
}

/////////////////////////////////////////////////////////////////
// Clears the LCD screen
//
MODULE_EXPORT void
roamer_clear (Driver *drvthis)
{
    PrivateData *p = drvthis->private_data;

    /* replace all chars in framebuf with spaces */
    memset(p->framebuf, ' ', (p->width * p->height));

    report(RPT_DEBUG, "%s: cleared framebuffer", drvthis->name);
}

//////////////////////////////////////////////////////////////////
// Flushes all output to the lcd...
//
MODULE_EXPORT void
roamer_flush (Driver *drvthis)
{
    PrivateData *p = drvthis->private_data;

    char out[ (p->width*p->height) + 2 ];

    report(RPT_DEBUG, "%s     Framebuf: '%.*s'", drvthis->name, p->width*p->height, p->framebuf);
    report(RPT_DEBUG, "%s Backingstore: '%.*s'", drvthis->name, p->width*p->height, p->backingstore);

    // skip over identical buffers
    if ( memcmp(p->framebuf, p->backingstore, p->width*p->height) == 0) {
        return;
    }

    snprintf(out, p->width*p->height+2, "M%.*s", p->width*p->height, p->framebuf);
    report(RPT_DEBUG, "%s writing: M%.*s", drvthis->name, p->width*p->height, p->framebuf);

    write(p->fd, out, sizeof(out)-1);

    memcpy(p->backingstore, p->framebuf, p->width * p->height);
    report(RPT_DEBUG, "%s: frame buffer flushed", drvthis->name);
}

/////////////////////////////////////////////////////////////////
// Prints a string on the lcd display, at position (x,y).  The
// upper-left is (1,1), and the lower right should be (20,4).
//
MODULE_EXPORT void
roamer_string (Driver *drvthis, int x, int y, const char string[])
{
    PrivateData *p = drvthis->private_data;
    int i;

    x--; y--; // Convert 1-based coords to 0-based...

    if ((y < 0) || (y >= p->height))
        return;

    for (i = 0; (string[i] != '\0') && (x < p->width); i++, x++) {
        if (x >= 0) // no write left of left border
            p->framebuf[(y * p->width) + x] = string[i];
    }
    report(RPT_DEBUG, "%s: printed string at (%d,%d)", drvthis->name, x, y);
}

/////////////////////////////////////////////////////////////////
// Prints a character on the lcd display, at position (x,y).  The
// upper-left is (1,1), and the lower right should be (20,4).
//
MODULE_EXPORT void
roamer_chr (Driver *drvthis, int x, int y, char c)
{
    PrivateData *p = drvthis->private_data;

    x--; y--; // Convert 1-based coords to 0-based...

    if (c == '#')
        c = 0xFF;

    if ((x >= 0) && (y >= 0) && (x < p->width) && (y < p->height))
        p->framebuf[(y * p->width) + x] = c;

    report(RPT_DEBUG, "%s: writing character %02X to position (%d,%d)", drvthis->name, c, x, y);
}

/////////////////////////////////////////////////////////////
// Does the heartbeat...
//
MODULE_EXPORT void
roamer_heartbeat (Driver *drvthis, int state)
{
    PrivateData *p = drvthis->private_data;

    static int timer = 0;
    static int saved_state = HEARTBEAT_ON;
    char out[3];
    char beat;

    if (state)
        saved_state = state;
    if (state == HEARTBEAT_ON) {
        beat = ((timer + 4) & 4) ? '1' : '0';
//        if (p->beat != beat) {
//            snprintf(out, 4, "L%c", beat);
//            write(p->fd, out, sizeof(out)-1);
//        }
//        p->beat = beat;

        roamer_chr(drvthis, p->width, 1, (beat == '1') ? '#' : ' ');
        roamer_flush(drvthis);
    }
    timer++;
    timer &= 0x0F;
}

/////////////////////////////////////////////////////////////////
// Sets the contrast
//
MODULE_EXPORT void
roamer_set_contrast (Driver *drvthis, int promille)
{
    //PrivateData *p = drvthis->private_data;

    report(RPT_DEBUG, "%s: contrast (%d)", drvthis->name, promille);
}

/////////////////////////////////////////////////////////////////
// Sets the backlight brightness
//
MODULE_EXPORT void
roamer_backlight (Driver *drvthis, int on)
{
    //PrivateData *p = drvthis->private_data;

    report(RPT_DEBUG, "%s: backlight (%s)", drvthis->name, (on) ? "ON" : "OFF");
}

